theory Typing
imports Syntax

begin

datatype Term = Term "ty"
types env = "(name * Term) set"

syntax
  var_term :: "name \<Rightarrow> ty \<Rightarrow> (name * Term)"  ("(_\<rhd>_)" [90,90] 90)
translations
  "(x\<rhd>\<tau>)" \<rightleftharpoons> "(x, Term \<tau>)"

syntax
  addbinding :: "(name * Term) set \<Rightarrow> name \<Rightarrow> ty \<Rightarrow> (name * Term) set"
    ("_','(_':_')" [90,0,0] 91)
translations
  "\<Gamma>,(x:\<tau>)" \<rightleftharpoons> "insert (x\<rhd>\<tau>) \<Gamma>"

constdefs wf_env :: "env \<Rightarrow> bool" ("_ \<Turnstile> OK" [50] 50)
  "wf_env \<Gamma> \<equiv> finite \<Gamma> \<and> (\<forall>x \<tau>1 \<tau>2. (x\<rhd>\<tau>1) \<in> \<Gamma> \<and> (x\<rhd>\<tau>2) \<in> \<Gamma> \<longrightarrow> \<tau>1 = \<tau>2)"

lemma wf_env_cons:
  assumes "\<Gamma> \<Turnstile> OK"
  and "\<not> \<Gamma> defines z"
  shows "\<Gamma>,(z:\<tau>) \<Turnstile> OK"
using prems by (auto simp add: dom_def wf_env_def)

consts Typing :: "(env * tm * ty) set"
syntax
  "_Typing" :: "env \<Rightarrow> tm \<Rightarrow> ty \<Rightarrow> bool"   ("_ \<turnstile> _ : _" [50,80,80] 50)
translations
  "\<Gamma> \<turnstile> e : \<tau>" \<rightleftharpoons> "(\<Gamma>, e, \<tau>) \<in> Typing"

inductive Typing
  intros
  t_var[intro!]  : "\<lbrakk> \<Gamma> \<Turnstile> OK; (x\<rhd>\<tau>) \<in> \<Gamma> \<rbrakk> \<Longrightarrow> \<Gamma> \<turnstile> TmFreeVar x : \<tau>"
  t_abs[intro!]  : "\<lbrakk> \<Gamma> \<Turnstile> OK; finite L;
                      \<forall>x. \<not> x free in e \<and> x \<notin> L \<and> \<not> \<Gamma> defines x \<longrightarrow>
                        \<Gamma>,(x:\<tau>1) \<turnstile> freshen_tm x e : \<tau>2 \<rbrakk> \<Longrightarrow>
                      \<Gamma> \<turnstile> \<lambda>[\<tau>1]. e : \<tau>1\<rightarrow>\<tau>2"
  t_app[intro!]   : "\<lbrakk> \<Gamma> \<turnstile> e1 : \<tau>1\<rightarrow>\<tau>2; \<Gamma> \<turnstile> e2 : \<tau>1 \<rbrakk> \<Longrightarrow>
                       \<Gamma> \<turnstile> e1\<cdot>e2 : \<tau>2"

inductive_cases typing_var_elim [elim]: "\<Gamma> \<turnstile> TmFreeVar x : \<tau>"
inductive_cases typing_abs_elim [elim]: "\<Gamma> \<turnstile> \<lambda>[\<tau>1]. e : \<tau>2"
inductive_cases typing_app_elim [elim]: "\<Gamma> \<turnstile> e1\<cdot>e2 : \<tau>"

lemma ty_wf_env:
  assumes "\<Gamma> \<turnstile> e : \<tau>"
  shows "\<Gamma> \<Turnstile> OK"
using prems by (induct, auto)

lemma ty_no_free_db:
  assumes "\<Gamma> \<turnstile> e : \<tau>"
  shows "free_db_tm e = []"
using prems proof induct
  case (t_abs L \<Gamma> \<tau>1 \<tau>2 e_b)
  then obtain z where "\<not> z free in e_b \<and> \<not> \<Gamma> defines z \<and> z \<notin> L"
    using wf_env_def dom_finite[of \<Gamma>] fresh_exists[of "set (free_tm e_b) \<union> dom \<Gamma> \<union> L"] by auto
  hence "set (free_db_tm (freshen_tm z e_b)) = {}" using t_abs by auto
  hence "set (free_db_tm e_b) - {0} = {}" by (auto simp add: vsubst_free_db)
  hence "set (free_db_tm e_b) \<subseteq> {0}" by auto
  hence "set (shift_db_list (free_db_tm e_b)) = {}" by (auto simp add: shift_many_sanity)
  hence "shift_db_list (free_db_tm e_b) = []" by auto
  thus ?case by auto
qed auto

lemma ty_same_subst:
  assumes "\<Gamma> \<turnstile> e : \<tau>"
  shows "vsubst_tm (TmFreeVar x) k e = e"
using prems by (auto simp only: ty_no_free_db no_free_db_vars_same_subst)

end
